from .document import document_schema
from .gallery import gallery_schema
from .presentation import presentation_schema
from .super_editor_document import super_editor_document_schema

__all__ = [
    document_schema,
    gallery_schema,
    presentation_schema,
    super_editor_document_schema,
]
